@@ -60,7 +60,7 @@ struct GsmXSMPClientPrivate
 
 enum {
         PROP_0,
-        PROP_ICE_CONNECTION,
+        PROP_ICE_CONNECTION
 };
 
 enum {
@@ -78,24 +78,37 @@ client_iochannel_watch (GIOChannel    *channel,
                         GIOCondition   condition,
                         GsmXSMPClient *client)
 {
+        gboolean keep_going;
 
+        g_object_ref (client);
         switch (IceProcessMessages (client->priv->ice_connection, NULL, NULL)) {
         case IceProcessMessagesSuccess:
-                return TRUE;
+                keep_going = TRUE;
+                break;
 
         case IceProcessMessagesIOError:
                 g_debug ("GsmXSMPClient: IceProcessMessagesIOError on '%s'", client->priv->description);
+                gsm_client_set_status (GSM_CLIENT (client), GSM_CLIENT_FAILED);
+                /* Emitting "disconnected" will eventually cause
+                 * IceCloseConnection() to be called.
+                 */
                 gsm_client_disconnected (GSM_CLIENT (client));
-                return FALSE;
+                keep_going = FALSE;
+                break;
 
         case IceProcessMessagesConnectionClosed:
                 g_debug ("GsmXSMPClient: IceProcessMessagesConnectionClosed on '%s'",
                          client->priv->description);
-                return FALSE;
+                client->priv->ice_connection = NULL;
+                keep_going = FALSE;
+                break;
 
         default:
                 g_assert_not_reached ();
         }
+        g_object_unref (client);
+
+        return keep_going;
 }
 
 /* Called if too much time passes between the initial connection and
@@ -108,6 +121,7 @@ _client_protocol_timeout (GsmXSMPClient *client)
                  client->priv->description,
                  IceConnectionStatus (client->priv->ice_connection));
 
+        gsm_client_set_status (GSM_CLIENT (client), GSM_CLIENT_FAILED);
         gsm_client_disconnected (GSM_CLIENT (client));
 
         return FALSE;
@@ -375,6 +389,7 @@ prop_to_command (SmProp *prop)
         return g_string_free (str, FALSE);
 }
 
+#if 0
 static char *
 xsmp_get_restart_command (GsmClient *client)
 {
@@ -402,12 +417,15 @@ xsmp_get_discard_command (GsmClient *client)
 
         return prop_to_command (prop);
 }
+#endif
 
 static void
 do_save_yourself (GsmXSMPClient *client,
                   int            save_type,
                   gboolean       forceful)
 {
+        g_assert (client->priv->conn != NULL);
+
         if (client->priv->next_save_yourself != -1) {
                 /* Either we're currently doing a shutdown and there's a checkpoint
                  * queued after it, or vice versa. Either way, the new SaveYourself
@@ -452,6 +470,7 @@ do_save_yourself (GsmXSMPClient *client,
         }
 }
 
+#if 0
 static void
 xsmp_save_yourself_phase2 (GsmClient *client)
 {
@@ -471,15 +490,26 @@ xsmp_interact (GsmClient *client)
 
         SmsInteract (xsmp->priv->conn);
 }
+#endif
 
-static void
-xsmp_cancel_end_session (GsmClient *client)
+static gboolean
+xsmp_cancel_end_session (GsmClient *client,
+                         GError   **error)
 {
         GsmXSMPClient *xsmp = (GsmXSMPClient *) client;
 
         g_debug ("GsmXSMPClient: xsmp_cancel_end_session ('%s')", xsmp->priv->description);
 
+        if (xsmp->priv->conn == NULL) {
+                g_set_error (error,
+                             GSM_CLIENT_ERROR,
+                             GSM_CLIENT_ERROR_NOT_REGISTERED,
+                             "Client is not registered");
+                return FALSE;
+        }
+
         SmsShutdownCancelled (xsmp->priv->conn);
+        return TRUE;
 }
 
 static gboolean
@@ -495,24 +525,44 @@ xsmp_stop (GsmClient *client,
         return TRUE;
 }
 
-static void
+static gboolean
 xsmp_query_end_session (GsmClient *client,
-                        guint      flags)
+                        guint      flags,
+                        GError   **error)
 {
         gboolean forceful;
 
+        if (GSM_XSMP_CLIENT (client)->priv->conn == NULL) {
+                g_set_error (error,
+                             GSM_CLIENT_ERROR,
+                             GSM_CLIENT_ERROR_NOT_REGISTERED,
+                             "Client is not registered");
+                return FALSE;
+        }
+
         forceful = (flags & GSM_CLIENT_END_SESSION_FLAG_FORCEFUL);
         do_save_yourself (GSM_XSMP_CLIENT (client), SmSaveGlobal, forceful);
+        return TRUE;
 }
 
-static void
+static gboolean
 xsmp_end_session (GsmClient *client,
-                  guint      flags)
+                  guint      flags,
+                  GError   **error)
 {
         gboolean forceful;
 
+        if (GSM_XSMP_CLIENT (client)->priv->conn == NULL) {
+                g_set_error (error,
+                             GSM_CLIENT_ERROR,
+                             GSM_CLIENT_ERROR_NOT_REGISTERED,
+                             "Client is not registered");
+                return FALSE;
+        }
+
         forceful = (flags & GSM_CLIENT_END_SESSION_FLAG_FORCEFUL);
         do_save_yourself (GSM_XSMP_CLIENT (client), SmSaveGlobal, forceful);
+        return TRUE;
 }
 
 static char *
@@ -575,27 +625,36 @@ gsm_xsmp_client_get_property (GObject    *object,
 }
 
 static void
-gsm_xsmp_client_finalize (GObject *object)
+gsm_xsmp_client_disconnect (GsmXSMPClient *client)
 {
-        GsmXSMPClient *client = (GsmXSMPClient *) object;
-
-        g_debug ("GsmXSMPClient: xsmp_finalize (%s)", client->priv->description);
-
         if (client->priv->watch_id > 0) {
                 g_source_remove (client->priv->watch_id);
         }
 
         if (client->priv->conn != NULL) {
                 SmsCleanUp (client->priv->conn);
-        } else {
+        }
+
+        if (client->priv->ice_connection != NULL) {
                 IceCloseConnection (client->priv->ice_connection);
         }
 
         if (client->priv->protocol_timeout > 0) {
                 g_source_remove (client->priv->protocol_timeout);
         }
+}
+
+static void
+gsm_xsmp_client_finalize (GObject *object)
+{
+        GsmXSMPClient *client = (GsmXSMPClient *) object;
+
+        g_debug ("GsmXSMPClient: xsmp_finalize (%s)", client->priv->description);
+        gsm_xsmp_client_disconnect (client);
 
         g_free (client->priv->description);
+        g_ptr_array_foreach (client->priv->props, (GFunc)SmFreeProperty, NULL);
+        g_ptr_array_free (client->priv->props, TRUE);
 
         G_OBJECT_CLASS (gsm_xsmp_client_parent_class)->finalize (object);
 }
@@ -812,12 +871,13 @@ register_client_callback (SmsConn    conn,
                 g_debug ("GsmXSMPClient: Sending initial SaveYourself");
                 SmsSaveYourself (conn, SmSaveLocal, False, SmInteractStyleNone, False);
                 client->priv->current_save_yourself = SmSaveLocal;
-
-                free (previous_id);
         }
 
         gsm_client_set_status (GSM_CLIENT (client), GSM_CLIENT_REGISTERED);
 
+        g_free (id);
+        free (previous_id);
+
         return TRUE;
 }
 
@@ -893,7 +953,21 @@ save_yourself_phase2_request_callback (SmsConn   conn,
         g_debug ("GsmXSMPClient: Client '%s' received SaveYourselfPhase2Request",
                  client->priv->description);
 
-        /* FIXME: change priority so this client runs at the end */
+        /* Treat this just like a SaveYourselfDone */
+        if (client->priv->current_save_yourself == SmSaveLocal) {
+                client->priv->current_save_yourself = -1;
+                SmsSaveComplete (client->priv->conn);
+        } else {
+                client->priv->current_save_yourself = -1;
+        }
+
+        /* this is a valid response to SaveYourself and therefore
+           may be a response to a QES or ES */
+        gdm_client_end_session_response (GSM_CLIENT (client),
+                                         TRUE,
+                                         NULL);
+
+        /* FIXME: change priority so this client runs at the end? */
 }
 
 static void
@@ -902,10 +976,12 @@ interact_request_callback (SmsConn   conn,
                            int       dialog_type)
 {
         GsmXSMPClient *client = manager_data;
+        gboolean       res;
+        GError        *error;
 
         g_debug ("GsmXSMPClient: Client '%s' received InteractRequest(%s)",
                  client->priv->description,
-                 dialog_type == SmInteractStyleAny ? "Any" : "Errors");
+                 dialog_type == SmDialogNormal ? "Dialog" : "Errors");
 
         gdm_client_end_session_response (GSM_CLIENT (client),
                                          FALSE,
@@ -917,7 +993,11 @@ interact_request_callback (SmsConn   conn,
            This grabbing is clearly bullshit and is not supported by
            the client spec or protocol spec.
         */
-        xsmp_cancel_end_session (GSM_CLIENT (client));
+        res = xsmp_cancel_end_session (GSM_CLIENT (client), &error);
+        if (! res) {
+                g_warning ("Unable to cancel end session: %s", error->message);
+                g_error_free (error);
+        }
 }
 
 static void
@@ -987,6 +1067,7 @@ close_connection_callback (SmsConn     conn,
         }
         SmFreeReasons (count, reason_msgs);
 
+        gsm_client_set_status (GSM_CLIENT (client), GSM_CLIENT_FINISHED);
         gsm_client_disconnected (GSM_CLIENT (client));
 }
 
